/**********************************************************************
 * FILE: mpi_pi_reduce.c
 * OTHER FILES: dboard.c
 * DESCRIPTION:
 *   MPI pi Calculation Example - C Version
 *   Collective Communication example:
 *   This program calculates pi using a "dartboard" algorithm.  See
 *   Fox et al.(1988) Solving Problems on Concurrent Processors, vol.1
 *   page 207.  All processes contribute to the calculation, with the
 *   master averaging the values for pi. This version uses mpc_reduce to
 *   collect results
 * AUTHOR: Blaise Barney. Adapted from Ros Leibensperger, Cornell Theory
 *   Center. Converted to MPI: George L. Gusciora, MHPCC (1/95)
* LAST REVISED: 06/13/13 Blaise Barney
* Traduction en français des commentaires et modification pour travailler
* à nombre total constant de lancers (IL 2017)
**********************************************************************/
#include "mpi.h"
#include <stdio.h>
#include <stdlib.h>

void srandom (unsigned seed);
double dboard (int darts);

#define DARTS  2000000    /* Nombre de lancer de fléchettes dans la cible par tour  */
#define ROUNDS 1000      /* Nombre de tours de lancers */
#define MASTER 0         /* Numéro du « Maître » */



int main (int argc, char *argv[])
{
double	homepi,     /* valeur de pi calculée par la tache courante */
	pisum,	     /* somme des valeurs de pi de chaque tache  */
	pi,	     /* moyenne de pi après chaque tour */
	avepi;	     /* average pi value for all iterations */
int	taskid,	     /* numéro de la tache  - et aussi graine du générateur aléatoire */
	numtasks,     /* nombre total de taches lancées */
	rc,           /* code de retour */
	i;
MPI_Status status;

/* on initialise les comm's et on récupère le nombre de taches et le numéro de notre tache */

MPI_Init(&argc,&argv);
MPI_Comm_size(MPI_COMM_WORLD,&numtasks);
MPI_Comm_rank(MPI_COMM_WORLD,&taskid);
printf ("la tache MPI %d à démarré ...\n", taskid);

/* la graine du générateur pseudo-aléatoire est fixée à la valeur du numéro de tache */
// srandom (taskid);
srand48(taskid);
printf(" alea %f tache %i \n",drand48(),taskid);
MPI_Barrier(MPI_COMM_WORLD);
avepi = 0;
for (i = 0; i < ROUNDS; i++) {
   /* toutes les tâches calculent pi par MC  */
   /* chaque tâche prend sa part de travail  DARTS/numtasks lancers */
   homepi = dboard(DARTS/numtasks);

   /* On utilise MPI_Reduce pour faire la somme des valeurs de homepi calculées par chaque tâche
    * la tâche Maître stockera la valeur de la somme (pisum)
    * - homepi est le buffer d'envoi
    * - pisum est le buffer de réception (utilisé seulement par la tâche qui reçoit)
    * - la taille du message est 1 réel double 
    * - MASTER est le numéro de la tâche qui recevra le résultat de l'opération de réduction 
    * - MPI_SUM est un drapeau prédéfini qui effectue la somme vectorielle. 
    * - MPI_COMM_WORLD est le groupe de communication défini au démarrage des tâches 
    */

   rc = MPI_Reduce(&homepi, &pisum, 1, MPI_DOUBLE, MPI_SUM, MASTER, MPI_COMM_WORLD);
   if (rc != MPI_SUCCESS)
      printf("%d: failure on mpc_reduce\n", taskid);

   /* la tâche maître calcule la moyenne pour cette itération pour les itérations totales*/
   if (taskid == MASTER) {
      pi = pisum/numtasks;
      avepi = ((avepi * i) + pi)/(i + 1);
      printf("   After %8d throws, average value of pi = %10.8f\n",
              (DARTS * (i + 1)),avepi);
   }
}
if (taskid == MASTER)
   printf ("\nReal value of PI: 3.1415926535897 \n");

MPI_Finalize();
return 0;
}

/**************************************************************************
* subroutine dboard
* DESCRIPTION:
*   Used in pi calculation example codes.
*   See mpi_pi_send.c and mpi_pi_reduce.c
*   Throw darts at board.  Done by generating random numbers
*   between 0 and 1 and converting them to values for x and y
*   coordinates and then testing to see if they "land" in
*   the circle."  If so, score is incremented.  After throwing the
*   specified number of darts, pi is calculated.  The computed value
*   of pi is returned as the value of this function, dboard.
*
*   Explanation of constants and variables used in this function:
*   darts       = number of throws at dartboard
*   score       = number of darts that hit circle
*   n           = index variable
*   r           = random number scaled between 0 and 1
*   x_coord     = x coordinate, between -1 and 1
*   x_sqr       = square of x coordinate
*   y_coord     = y coordinate, between -1 and 1
*   y_sqr       = square of y coordinate
*   pi          = computed value of pi
****************************************************************************/

double dboard(int darts)
{
#define sqr(x)	((x)*(x))
long random(void);
double x_coord, y_coord, pi, r;
int score, n;
unsigned int cconst;  /* must be 4-bytes in size */
/*************************************************************************
 * The cconst variable must be 4 bytes. We check this and bail if it is
 * not the right size
 ************************************************************************/
if (sizeof(cconst) != 4) {
   printf("Wrong data size for cconst variable in dboard routine!\n");
   printf("See comments in source file. Quitting.\n");
   exit(1);
   }
   /* 2 bit shifted to MAX_RAND later used to scale random number between 0 and 1 */
   cconst = 2 << (31 - 1);
   score = 0;

   /* "throw darts at board" */
   for (n = 1; n <= darts; n++)  {
      /* generate random numbers for x and y coordinates */
      // r = (double)random()/cconst;
      r=drand48();
	x_coord = (2.0 * r) - 1.0;
      //r = (double)random()/cconst;
      r=drand48();
	y_coord = (2.0 * r) - 1.0;

      /* if dart lands in circle, increment score */
      if ((sqr(x_coord) + sqr(y_coord)) <= 1.0)
           score++;
      }
/* calculate pi */
pi = 4.0 * (double)score/(double)darts;
return(pi);
}

